<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 08/26/2003 *
 * ===============                              Last change: 11/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : mysql-manager.php                                *
 * -------------------------------------------------------------------- *
 * Short description : All MySQL-related functions                      *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Alle MySQL-Relevanten Funktionen                 *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * Needs to be in all Files and every File needs "svn propset           *
 * svn:keywords Date Revision" (autoprobset!) at least!!!!!!            *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// "Getter" for module title
function getModuleTitle ($mod) {
	// Init variables
	$title = '';
	$result = false;

	// Is the script installed?
	if (isInstalled()) {
		// Check if cache is valid
		if ((isExtensionInstalledAndNewer('cache', '0.1.2')) && (isset($GLOBALS['cache_array']['modules']['module'])) && (in_array($mod, $GLOBALS['cache_array']['modules']['module']))) {
			// Load from cache
			$title = $GLOBALS['cache_array']['modules']['title'][$mod];

			// Update cache hits
			incrementStatsEntry('cache_hits');
		} elseif (!isExtensionActive('cache')) {
			// Load from database
			$result = SQL_QUERY_ESC("SELECT `title` FROM `{?_MYSQL_PREFIX?}_mod_reg` WHERE `module`='%s' LIMIT 1",
				array($mod), __FUNCTION__, __LINE__);

			// Is the entry there?
			if (SQL_NUMROWS($result)) {
				// Get the title from database
				list($title) = SQL_FETCHROW($result);
			} // END - if

			// Free the result
			SQL_FREERESULT($result);
		}
	} // END - if

	// Trim name
	$title = trim($title);

	// Still no luck or empty title?
	if (empty($title)) {
		// No name found
		$title = sprintf("%s (%s)", getMessage('LANG_UNKNOWN_MODULE'), $mod);
		if (SQL_NUMROWS($result) == 0) {
			// Add module to database
			$dummy = checkModulePermissions($mod);
		} // END - if
	} // END - if

	// Return name
	return $title;
}

// "Getter" for module description
// @TODO Can we cache this?
function getModuleDescription ($mode, $what, $column = 'what') {
	// Fix empty 'what'
	if (empty($what)) {
		$what = getConfig('index_home');
	} // END - if

	// Default is not found
	$ret = '??? (' . $what . ')';

	// Look for title
	$result = SQL_QUERY_ESC("SELECT `title` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE %s='%s' LIMIT 1",
		array($mode, $column, $what), __FUNCTION__, __LINE__);

	// Is there an entry?
	if (SQL_NUMROWS($result) == 1) {
		// Fetch the title
		list($ret) = SQL_FETCHROW($result);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return it
	return $ret;
}

// Check validity of a given module name (no file extension)
function checkModulePermissions ($mod = '') {
	// Is it empty (default), then take the current one
	if (empty($mod)) $mod = getModule();

	// Do we have cache?
	if (isset($GLOBALS['module_status'][$mod])) {
		// Then use it
		return $GLOBALS['module_status'][$mod];
	} // END - if

	// Filter module name (names with low chars and underlines are fine!)
	$mod = preg_replace('/[^a-z_]/', '', $mod);

	// Check for prefix is a extension...
	$modSplit = explode('_', $mod);
	$extension = ''; $mod_chk = $mod;
	//* DEBUG: */ outputHtml(__LINE__."*".count($modSplit)."*/".$mod."*<br />");
	if (count($modSplit) == 2) {
		// Okay, there is a seperator (_) in the name so is the first part a module?
		//* DEBUG: */ outputHtml(__LINE__."*".$modSplit[0]."*<br />");
		if (isExtensionActive($modSplit[0])) {
			// The prefix is an extension's name, so let's set it
			$extension = $modSplit[0]; $mod = $modSplit[1];
		} // END - if
	} // END - if

	// Major error in module registry is the default
	$ret = 'major';

	// Check if script is installed if not return a 'done' to prevent some errors
	if ((isInstallationPhase()) || (!isAdminRegistered())) {
		// Not installed or no admin registered or in installation phase
		return 'done';
	} // END - if

	// Init variables
	$locked = 'Y';
	$hidden = 'N';
	$admin  = 'N';
	$mem    = 'N';
	$found  = false;

	// Check if cache is latest version
	if (isExtensionInstalledAndNewer('cache', '0.1.2')) {
		// Is the cache there?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using cache.');
		if (isset($GLOBALS['cache_array']['modules']['locked'][$mod_chk])) {
			// Check cache
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Cache found.');
			$locked = $GLOBALS['cache_array']['modules']['locked'][$mod_chk];
			$hidden = $GLOBALS['cache_array']['modules']['hidden'][$mod_chk];
			$admin  = $GLOBALS['cache_array']['modules']['admin_only'][$mod_chk];
			$mem    = $GLOBALS['cache_array']['modules']['mem_only'][$mod_chk];

			// Update cache hits
			incrementStatsEntry('cache_hits');
			$found = true;
		} else {
			// No, then we have to update it!
			$ret = 'cache_miss';
		}
	} elseif (!isExtensionActive('cache')) {
		// Check for module in database
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using database.');
		$result = SQL_QUERY_ESC("SELECT `locked`, `hidden`, `admin_only`, `mem_only` FROM `{?_MYSQL_PREFIX?}_mod_reg` WHERE `module`='%s' LIMIT 1",
			array($mod_chk), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Read data
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Entry found.');
			list($locked, $hidden, $admin, $mem) = SQL_FETCHROW($result);
			$found = true;
		} elseif (isDebugModeEnabled()) {
			// Debug message only in debug-mode...
			logDebugMessage(__FUNCTION__, __LINE__, 'Module ' . $mod_chk . ' not found!');
		}

		// Free result
		SQL_FREERESULT($result);
	}
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'ret=' . $ret);

	// Is the module found?
	if ($found === true) {
		// Check returned values against current access permissions
		//
		//  Admin access            ----- Guest access -----           --- Guest   or   member? ---
		if ((isAdmin()) || (($locked != 'Y') && ($admin != 'Y') && (($mem != 'Y') || (isMember())))) {
			// If you are admin you are welcome for everything!
			$ret = 'done';
		} elseif ($locked == 'Y') {
			// Module is locked
			$ret = 'locked';
		} elseif (($mem == 'Y') && (!isMember())) {
			// You have to login first!
			$ret = 'mem_only';
		} elseif (($admin == 'Y') && (!isAdmin())) {
			// Only the Admin is allowed to enter this module!
			$ret = 'admin_only';
		} else {
			// @TODO Nothing helped???
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("ret=%s,locked=%s,admin=%s,mem=%s",
				$ret,
				$locked,
				$admin,
				$mem
			));
		}
	} // END - if

	// Still no luck or not found?
	if (($found === false) && (!isExtensionActive('cache')) && ($ret != 'done'))  {
		//              ----- Legacy module -----                                               ---- Module in base folder  ----                       --- Module with extension's name ---
		if ((isIncludeReadable(sprintf("inc/modules/%s.php", $mod))) || (isIncludeReadable(sprintf("%s.php", $mod))) || (isIncludeReadable(sprintf("%s/%s.php", $extension, $mod)))) {
			// Data is missing so we add it
			if (isExtensionInstalledAndNewer('sql_patches', '0.3.6')) {
				// Since 0.3.6 we have a has_menu column, this took me a half hour
				// to find a loop here... *sigh*
				SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_mod_reg`
(`module`, `locked`, `hidden`, `mem_only`, `admin_only`, `has_menu`) VALUES
('%s','Y','N','N','N','N')", array($mod_chk), __FUNCTION__, __LINE__);
			} else {
				// Wrong/missing sql_patches!
				SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_mod_reg`
(`module`, `locked`, `hidden`, `mem_only`, `admin_only`) VALUES
('%s','Y','N','N','N')", array($mod_chk), __FUNCTION__, __LINE__);
			}

			// Everthing is fine?
			if (SQL_AFFECTEDROWS() < 1) {
				// Something bad happend!
				return 'major';
			} // END - if

			// Destroy cache here
			// @TODO Rewrite this to a filter
			if (getOutputMode() == 0) rebuildCacheFile('modules', 'modules');

			// And reload data
			unset($GLOBALS['module_status'][$mod]);
			$ret = checkModulePermissions($mod_chk);
		} else {
			// Module not found we don't add it to the database
			$ret = '404';
		}
	} elseif (($ret == 'cache_miss') && (getOutputMode() == 0)) {
		// Rebuild the cache files
		rebuildCacheFile('modules', 'modules');
	} elseif ($found === false) {
		// Problem with module detected
		logDebugMessage(__FUNCTION__, __LINE__, sprintf("Problem in module %s detected. ret=%s, locked=%s, hidden=%s, mem=%s, admin=%s, output_mode=%s",
			$mod,
			$ret,
			$locked,
			$hidden,
			$mem,
			$admin,
			getOutputMode()
		));
	}

	// Return the value
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'ret=' . $ret);
	$GLOBALS['module_status'][$mod] = $ret;
	return $ret;
}

// Add menu description pending on given file name (without path!)
function addMenuDescription ($accessLevel, $FQFN, $return = false, $output = true) {
	// Use only filename of the FQFN...
	$file = basename($FQFN);

	// Init variables
	$LINK_ADD = '';
	$OUT = '';
	$AND = '';

	// First we have to do some analysis...
	if (substr($file, 0, 7) == 'action-') {
		// This is an action file!
		$type = 'action';
		$search = substr($file, 7);
		switch ($accessLevel) {
			case 'admin':
				$modCheck = 'admin';
				break;

			case 'sponsor':
			case 'guest':
			case 'member':
				$modCheck = getModule();
				break;
		}
		$AND = " AND (`what`='' OR `what` IS NULL)";
	} elseif (substr($file, 0, 5) == 'what-') {
		// This is an admin what file!
		$type = 'what';
		$search = substr($file, 5);
		$AND = '';
		switch ($accessLevel) {
			case 'admin':
				$modCheck = 'admin';
				break;

			case 'guest':
			case 'member':
				$modCheck = getModule();
				if (!isAdmin()) {
					$AND = " AND `visible`='Y' AND `locked`='N'";
				}
				break;
		}

		$dummy = substr($search, 0, -4);
		$AND .= " AND `action`='".getModeAction($accessLevel, $dummy)."'";
	} elseif (($accessLevel == 'sponsor') || ($accessLevel == "engine")) {
		// Sponsor / engine menu
		$type     = 'what';
		$search   = $file;
		$modCheck = getModule();
		$AND      = '';
	} else {
		// Other
		$type     = 'menu';
		$search   = $file;
		$modCheck = getModule();
		$AND      = '';
	}

	// Begin the navigation line
	if ((!isset($GLOBALS['nav_depth'])) && ($return === false)) {
		$GLOBALS['nav_depth'] = 0;
		$prefix = "<div class=\"you_are_here\">{--YOU_ARE_HERE--}&nbsp;<strong><a class=\"you_are_here\" href=\"{?URL?}/modules.php?module=".getModule().$LINK_ADD."\">Home</a></strong>";
	} else {
		if ($return === false) $GLOBALS['nav_depth']++;
		$prefix = '';
	}

	$prefix .= '&nbsp;-&gt;&nbsp;';

	// We need to remove .php and the end
	if (substr($search, -4, 4) == '.php') {
		// Remove the .php
		$search = substr($search, 0, -4);
	} // END - i

	// Get the title from menu
	$result = SQL_QUERY_ESC("SELECT title FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE %s='%s' ".$AND." LIMIT 1",
	array($accessLevel, $type, $search), __FUNCTION__, __LINE__);

	// Menu found?
	if (SQL_NUMROWS($result) == 1) {
		// Load title
		list($ret) = SQL_FETCHROW($result);

		// Shall we return it?
		if ($return === true) {
			// Return title
			return $ret;
		} elseif (((isExtensionInstalledAndNewer('sql_patches', '0.2.3')) && (getConfig('youre_here') == 'Y')) || ((isAdmin()) && ($modCheck == 'admin'))) {
			// Output HTML code
			$OUT = $prefix . "<strong><a class=\"you_are_here\" href=\"{?URL?}/modules.php?module=" . $modCheck . '&amp;' . $type . '=' . $search.$LINK_ADD . "\">" . $ret . "</a></strong>\n";

			// Can we close the you-are-here navigation?
			//* DEBUG: */ outputHtml(__LINE__."*".$type.'/'.getWhat()."*<br />");
			if (($type == 'what') || (($type == 'action') && ((!isWhatSet()) || (getWhat() == 'overview')))) {
				//* DEBUG: */ outputHtml(__LINE__.'+'.$type."+<br />");
				// Add closing div and br-tag
				$OUT .= "</div><br />\n";
				$GLOBALS['nav_depth'] = 0;

				// Run the filter chain
				$ret = runFilterChain('post_youhere_line', array('access_level' => $accessLevel, 'type' => $type, 'content' => ""));
				$OUT .= $ret['content'];
			} // END - if
		}
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return or output HTML code?
	if ($output) {
		// Output HTML code here
		outputHtml($OUT);
	} else {
		// Return HTML code
		return $OUT;
	}
}

// Adds a menu (mode = guest/member/admin/sponsor) to output
function addMenu ($mode, $action, $what) {
	// Init some variables
	$main_cnt = 0;
	$AND = '';

	// is the menu action valid?
	if (!isMenuActionValid($mode, $action, $what, true)) {
		return getCode('MENU_NOT_VALID');
	} // END - if

	// Non-admin shall not see all menus
	if (!isAdmin()) {
		$AND = " AND `visible`='Y' AND `locked`='N'";
	} // END - if

	// Load SQL data and add the menu to the output stream...
	$result_main = SQL_QUERY_ESC("SELECT `title`, `action` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE (`what`='' OR `what` IS NULL)".$AND." ORDER BY `sort` ASC",
		array($mode), __FUNCTION__, __LINE__);

	//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.':'.getWhat()."*<br />");
	if (SQL_NUMROWS($result_main) > 0) {
		// There are menus available, so we simply display them... :)
		$GLOBALS['rows'] = '';
		while ($content = SQL_FETCHARRAY($result_main)) {
			//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.'/'.$content['action'].':'.getWhat()."*<br />");
			// Init variables
			enableBlockMode(false);
			$action = $content['action'];

			// Load menu header template
			$GLOBALS['rows'] .= loadTemplate($mode . '_menu_title', true, $content);

			// Sub menu
			$result_sub = SQL_QUERY_ESC("SELECT `title` AS sub_title, `what` AS sub_what FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND `what` != '' AND `what` IS NOT NULL ".$AND." ORDER BY `sort`",
				array($mode, $content['action']), __FUNCTION__, __LINE__);

			// Get number of rows
			$totalWhats = SQL_NUMROWS($result_sub);

			// Do we have some entries?
			if ($totalWhats > 0) {
				// Init counter
				$cnt = 0;

				// Load all sub menus
				while ($content2 = SQL_FETCHARRAY($result_sub)) {
					// Merge both arrays in one
					$content = merge_array($content, $content2);

					// Init content
					$OUT = '';

					// Full file name for checking menu
					//* DEBUG: */ outputHtml(__LINE__.":!!!!".$content['sub_what']."!!!<br />");
					$inc = sprintf("inc/modules/%s/what-%s.php", $mode, $content['sub_what']);
					if (isIncludeReadable($inc)) {
						// Mark currently selected menu - open
						if ((!empty($what)) && (($what == $content['sub_what']))) {
							$OUT = "<strong>";
						} // END - if

						// Navigation link
						$OUT .= "<a name=\"menu\" class=\"menu_blur\" href=\"{?URL?}/modules.php?module=".getModule()."&amp;what=".$content['sub_what']."\" target=\"_self\">";
					} else {
						// Not found! - open
						$OUT .= "<em style=\"cursor:help\" class=\"admin_note\" title=\"{--MENU_WHAT_404--}\">";
					}

					// Menu title
					$OUT .= getConfig('menu_blur_spacer') . $content['sub_title'];

					if (isIncludeReadable($inc)) {
						$OUT .= "</a>";

						// Mark currently selected menu - close
						if ((!empty($what)) && (($what == $content['sub_what']))) {
							$OUT .= "</strong>";
						} // END - if
					} else {
						// Not found! - close
						$OUT .= "</em>";
					}

					// Cunt it up
					$cnt++;

					// Rewrite array
					$content =  array(
						'menu' => $OUT,
						'what' => $content['sub_what']
					);

					// Add regular menu row or bottom row?
					if ($cnt < $totalWhats) {
						$GLOBALS['rows'] .= loadTemplate($mode . '_menu_row', true, $content);
					} else {
						$GLOBALS['rows'] .= loadTemplate($mode . '_menu_bottom', true, $content);
					}
				}
			} else {
				// This is a menu block... ;-)
				enableBlockMode();

				// Load menu block
				$INC = sprintf("inc/modules/%s/action-%s.php", $mode, $content['action']);
				if (isFileReadable($INC)) {
					// Load include file
					if ((!isExtensionActive($content['action'])) || ($content['action'] == 'online')) $GLOBALS['rows'] .= loadTemplate('menu_what_begin', true, $mode);
					//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.'/'.$content['action'].'/'.getWhat()."*<br />");
					loadInclude($INC);
					//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.'/'.$content['action'].'/'.getWhat()."*<br />");
					if ((!isExtensionActive($content['action'])) || ($content['action'] == 'online')) $GLOBALS['rows'] .= loadTemplate('menu_what_end', true, $mode);
				}
				//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.'/'.$content['action'].'/'.$content['sub_what'].':'.getWhat()."*<br />");
			}

			$main_cnt++;

			//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.':'.getWhat()."*<br />");
			if (SQL_NUMROWS($result_main) > $main_cnt) {
				// Add seperator
				$GLOBALS['rows'] .= loadTemplate('menu_seperator', true, $mode);

				// Should we display adverts in this menu?
				if ((isExtensionInstalledAndNewer('menu', '0.0.1')) && (getConfig($mode . '_menu_advert_enabled') == 'Y') && ($action != 'admin')) {
					// Display advert template
					$GLOBALS['rows'] .= loadTemplate('menu_' . $mode . '_advert_' . $action, true);

					// Add seperator again
					$GLOBALS['rows'] .= loadTemplate('menu_seperator', true, $mode);
				} // END - if
			} // END - if
		} // END - while

		// Free memory
		SQL_FREERESULT($result_main);

		// Should we display adverts in this menu?
		if ((isExtensionInstalledAndNewer('menu', '0.0.1')) && (getConfig($mode . '_menu_advert_enabled') == 'Y')) {
			// Add seperator again
			$GLOBALS['rows'] .= loadTemplate('menu_seperator', true, $mode);

			// Display advert template
			$GLOBALS['rows'] .= loadTemplate('menu_' . $mode . '_advert_end', true);
		} // END - if

		// Prepare data
		$content = array(
			'rows' => $GLOBALS['rows'],
			'mode' => $mode
		);

		// Load main template
		//* DEBUG: */ outputHtml(__LINE__.'/'.$main_cnt.'/'.$content['action'].'/'.$content['sub_what'].':'.getWhat()."*<br />");
		loadTemplate('menu_table', false, $content);
	} // END - if
}

// Checks wether the current user is a member
function isMember () {
	// By default no member
	$ret = false;

	// Fix missing 'last_online' array, damn stupid code :(((
	// @TODO Try to rewrite this to one or more functions
	if ((!isset($GLOBALS['last_online'])) || (!is_array($GLOBALS['last_online']))) $GLOBALS['last_online'] = array();

	// is the cache entry there?
	if (isset($GLOBALS['is_member'])) {
		// Then return it
		return $GLOBALS['is_member'];
	} // END - if

	// Init global 'status'
	$GLOBALS['status'] = false;

	// Fix "deleted" cookies first
	fixDeletedCookies(array('userid', 'u_hash'));

	// Are cookies set?
	if ((isUserIdSet()) && (isSessionVariableSet('u_hash'))) {
		// Cookies are set with values, but are they valid?
		$result = SQL_QUERY_ESC("SELECT `password`, `status`, `last_module`, `last_online` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
			array(getUserId()), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Load data from cookies
			list($password, $GLOBALS['status'], $mod, $onl) = SQL_FETCHROW($result);

			// Validate password by created the difference of it and the secret key
			$valPass = generatePassString($password);

			// Transfer last module and online time
			if ((!empty($mod)) && (empty($GLOBALS['last_online']['module']))) {
				// @TODO Try to rewrite this to one or more functions
				$GLOBALS['last_online']['module'] = $mod;
				$GLOBALS['last_online']['online'] = $onl;
			} // END - if

			// So did we now have valid data and an unlocked user?
			if (($GLOBALS['status'] == 'CONFIRMED') && ($valPass == getSession('u_hash'))) {
				// Account is confirmed and all cookie data is valid so he is definely logged in! :-)
				$ret = true;
			} else {
				// Maybe got locked etc.
				//* DEBUG: */ outputHtml(__LINE__."!!!<br />");
				destroyUserSession();
			}
		} else {
			// Cookie data is invalid!
			//* DEBUG: */ outputHtml(__LINE__."***<br />");
			destroyUserSession();
		}

		// Free memory
		SQL_FREERESULT($result);
	} else {
		// Cookie data is invalid!
		//* DEBUG: */ outputHtml(__LINE__."///<br />");
		destroyUserSession();
	}

	// Cache status
	$GLOBALS['is_member'] = $ret;

	// Return status
	return $ret;
}

// This patched function will reduce many SELECT queries for the specified or current admin login
function isAdmin ($admin = '') {
	// Init variables
	$ret = false; $passCookie = ''; $valPass = '';
	//* DEBUG: */ outputHtml(__LINE__."ADMIN:".$admin."<br />");

	// If admin login is not given take current from cookies...
	if ((empty($admin)) && (isSessionVariableSet('admin_login')) && (isSessionVariableSet('admin_md5'))) {
		// Get admin login and password from session/cookies
		$admin = getSession('admin_login');
		$passCookie = getSession('admin_md5');
	} // END - if
	//* DEBUG: */ outputHtml(__LINE__."ADMIN:".$admin.'/'.$passCookie."<br />");

	// Search in array for entry
	if (isset($GLOBALS['admin_hash'])) {
		// Use cached string
		$valPass = $GLOBALS['admin_hash'];
	} elseif ((!empty($passCookie)) && (isAdminHashSet($admin) === true) && (!empty($admin))) {
		// Login data is valid or not?
		$valPass = generatePassString(getAdminHash($admin));

		// Cache it away
		$GLOBALS['admin_hash'] = $valPass;

		// Count cache hits
		incrementStatsEntry('cache_hits');
	} elseif ((!empty($admin)) && ((!isExtensionActive('cache'))) || (isAdminHashSet($admin) === false)) {
		// Search for admin
		$result = SQL_QUERY_ESC("SELECT HIGH_PRIORITY `password` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `login`='%s' LIMIT 1",
			array($admin), __FUNCTION__, __LINE__);

		// Is he admin?
		$passDB = '';
		if (SQL_NUMROWS($result) == 1) {
			// Admin login was found so let's load password from DB
			list($passDB) = SQL_FETCHROW($result);

			// Temporary cache it
			setAdminHash($admin, $passDB);

			// Generate password hash
			$valPass = generatePassString($passDB);
		} // END - if

		// Free memory
		SQL_FREERESULT($result);
	}

	if (!empty($valPass)) {
		// Check if password is valid
		//* DEBUG: */ print(__FUNCTION__ . ':' . $valPass . '/' . $passCookie . '<br />');
		$ret = (($valPass == $passCookie) || ((strlen($valPass) == 32) && ($valPass == md5($passCookie))) || (($valPass == '*FAILED*') && (!isExtensionActive('cache'))));
	} // END - if

	// Return result of comparision
	return $ret;
}

// Generates a list of "max receiveable emails per day"
function addMaxReceiveList ($mode, $default = '', $return = false) {
	$OUT = '';
	$result = false;

	switch ($mode) {
		case 'guest':
			// Guests (in the registration form) are not allowed to select 0 mails per day.
			$result = SQL_QUERY("SELECT value, comment FROM `{?_MYSQL_PREFIX?}_max_receive` WHERE value > 0 ORDER BY value",
			__FUNCTION__, __LINE__);
			break;

		case 'member':
			// Members are allowed to set to zero mails per day (we will change this soon!)
			$result = SQL_QUERY("SELECT value, comment FROM `{?_MYSQL_PREFIX?}_max_receive` ORDER BY value",
			__FUNCTION__, __LINE__);
			break;

		default: // Invalid!
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid mode %s detected.", $mode));
			break;
	}

	// Some entries are found?
	if (SQL_NUMROWS($result) > 0) {
		$OUT = '';
		while ($content = SQL_FETCHARRAY($result)) {
			$OUT .= "      <option value=\"".$content['value']."\"";
			if (postRequestElement('max_mails') == $content['value']) $OUT .= ' selected="selected"';
			$OUT .= ">".$content['value']." {--PER_DAY--}";
			if (!empty($content['comment'])) $OUT .= " (".$content['comment'].')';
			$OUT .= "</option>\n";
		}

		// Load template
		$OUT = loadTemplate(($mode . '_receive_table'), true, $OUT);
	} else {
		// Maybe the admin has to setup some maximum values?
		debug_report_bug('Nothing is being done here?');
	}

	// Free result
	SQL_FREERESULT($result);

	if ($return === true) {
		// Return generated HTML code
		return $OUT;
	} else {
		// Output directly (default)
		outputHtml($OUT);
	}
}

// Checks wether the given email address is used.
function isEmailTaken ($email) {
	// Query the database
	$result = SQL_QUERY_ESC("SELECT `userid` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE email LIKE '{PER}%s{PER}' LIMIT 1",
		array($email), __FUNCTION__, __LINE__);

	// Is the email there?
	$ret = (SQL_NUMROWS($result) == 1);

	// Free the result
	SQL_FREERESULT($result);

	// Return result
	return $ret;
}

// Validate the given menu action
function isMenuActionValid ($mode, $action, $what, $updateEntry=false) {
	// Is the cache entry there and we shall not update?
	if ((isset($GLOBALS['action_valid'][$mode][$action][$what])) && ($updateEntry === false)) {
		// Count cache hit
		incrementStatsEntry('cache_hits');

		// Then use this cache
		return $GLOBALS['action_valid'][$mode][$action][$what];
	} // END - if

	// By default nothing is valid
	$ret = false;

	// Look in all menus or only unlocked
	$add = '';
	if ((!isAdmin()) && ($mode != 'admin')) $add = " AND `locked`='N'";

	//* DEBUG: */ outputHtml(__LINE__.':'.$mode.'/'.$action.'/'.$what."*<br />");
	if (($mode != 'admin') && ($updateEntry === true)) {
		// Update guest or member menu
		$sql = SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET counter=counter+1 WHERE `action`='%s' AND `what`='%s'".$add." LIMIT 1",
			array($mode, $action, $what), __FUNCTION__, __LINE__, false);
	} elseif (($what != 'overview') && (!empty($what))) {
		// Other actions
		$sql = SQL_QUERY_ESC("SELECT `id`, `what` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND `what`='%s'".$add." ORDER BY `action` DESC LIMIT 1",
			array($mode, $action, $what), __FUNCTION__, __LINE__, false);
	} else {
		// Admin login overview
		$sql = SQL_QUERY_ESC("SELECT `id`, `what` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND (`what`='' OR `what` IS NULL)".$add." ORDER BY `action` DESC LIMIT 1",
			array($mode, $action), __FUNCTION__, __LINE__, false);
	}

	// Run SQL command
	$result = SQL_QUERY($sql, __FUNCTION__, __LINE__);
	if ($updateEntry === true) {
		// Check updated/affected rows
		$ret = (SQL_AFFECTEDROWS() == 1);
	} else {
		// Check found rows
		$ret = (SQL_NUMROWS($result) == 1);
	}

	// Free memory
	SQL_FREERESULT($result);

	// Set cache entry
	$GLOBALS['action_valid'][$mode][$action][$what] = $ret;

	// Return result
	return $ret;
}

//
function sendModeMails ($mod, $modes) {
	// Load hash
	$result_main = SQL_QUERY_ESC("SELECT `password` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s AND `status`='CONFIRMED' LIMIT 1",
		array(getUserId()), __FUNCTION__, __LINE__);
	if (SQL_NUMROWS($result_main) == 1) {
		// Load hash from database
		list($hashDB) = SQL_FETCHROW($result_main);

		// Extract salt from cookie
		$salt = substr(getSession('u_hash'), 0, -40);

		// Now let's compare passwords
		$hash = generatePassString($hashDB);
		if (($hash == getSession('u_hash')) || (postRequestElement('pass1') == postRequestElement('pass2'))) {
			// Load user's data               0        1        2         3        4      5    6      7
			$result = SQL_QUERY_ESC("SELECT gender, surname, family, street_nr, country, zip, city, email FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s AND password='%s' LIMIT 1",
				array(getUserId(), $hashDB), __FUNCTION__, __LINE__);
			if (SQL_NUMROWS($result) == 1) {
				// Load the data
				$content = SQL_FETCHARRAY($result, 0, false);

				// Free result
				SQL_FREERESULT($result);

				// Translate gender
				$content['gender'] = translateGender($content['gender']);

				// Clear/init the content variable
				$content['info'] = '';

				switch ($mod) {
					case 'mydata':
						foreach ($modes as $mode) {
							switch ($mode) {
								case 'normal': break; // Do not add any special lines
								case 'email': // Email was changed!
									$content['message'] = getMessage('MEMBER_CHANGED_EMAIL').": ".postRequestElement('old_email')."\n";
									break;

								case 'pass': // Password was changed
									$content['message'] = getMessage('MEMBER_CHANGED_PASS')."\n";
									break;

								default:
									logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown mode %s detected.", $mode));
									$content['message'] = getMessage('MEMBER_UNKNOWN_MODE').": ".$mode."\n\n";
									break;
							} // END - switch
						} // END - if

						if (isExtensionActive('country')) {
							// Replace code with description
							$content['country'] = generateCountryInfo(postRequestElement('country_code'));
						} // END - if

						// Merge content with data from POST
						$content = merge_array($content, postRequestArray());

						// Load template
						$message = loadEmailTemplate('member_mydata_notify', $content, getUserId());

						if (getConfig('admin_notify') == 'Y') {
							// The admin needs to be notified about a profile change
							$message_admin = 'admin_mydata_notify';
							$sub_adm   = getMessage('ADMIN_CHANGED_DATA');
						} else {
							// No mail to admin
							$message_admin = '';
							$sub_adm   = '';
						}

						// Set subject lines
						$sub_mem = getMessage('MEMBER_CHANGED_DATA');

						// Output success message
						$content = "<span class=\"member_done\">{--MYDATA_MAIL_SENT--}</span>";
						break;

					default: // Unsupported module!
						logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unsupported module %s detected.", $mod));
						$content = "<span class=\"member_failed\">{--UNKNOWN_MODULE--}</span>";
						break;
				} // END - switch
			} else {
				// Could not load profile data
				$content = "<span class=\"member_failed\">{--MEMBER_CANNOT_LOAD_PROFILE--}</span>";
			}
		} else {
			// Passwords mismatch
			$content = "<span class=\"member_failed\">{--MEMBER_PASSWORD_ERROR--}</span>";
		}
	} else {
		// Could not load profile
		$content = "<span class=\"member_failed\">{--MEMBER_CANNOT_LOAD_PROFILE--}</span>";
	}

	// Send email to user if required
	if ((!empty($sub_mem)) && (!empty($message))) {
		// Send member mail
		sendEmail($content['email'], $sub_mem, $message);
	} // END - if

	// Send only if no other error has occured
	if (empty($content)) {
		if ((!empty($sub_adm)) && (!empty($message_admin))) {
			// Send admin mail
			sendAdminNotification($sub_adm, $message_admin, $content, getUserId());
		} elseif (getConfig('admin_notify') == 'Y') {
			// Cannot send mails to admin!
			$content = getMessage('CANNOT_SEND_ADMIN_MAILS');
		} else {
			// No mail to admin
			$content = "<span class=\"member_done\">{--MYDATA_MAIL_SENT--}</span>";
		}
	} // END - if

	// Load template
	loadTemplate('admin_settings_saved', false, $content);
}

// Get action value from mode (admin/guest/member) and what-value
function getModeAction ($mode, $what) {
	// Init status
	$ret = '';

	//* DEBUG: */ outputHtml(__LINE__.'='.$mode.'/'.$what.'/'.getAction()."=<br />");
	if ((empty($what)) && ($mode != 'admin')) {
		$what = getConfig('index_home');
	} // END - if

	if ($mode == 'admin') {
		// Action value for admin area
		if (isGetRequestElementSet('action')) {
			// Use from request!
			return getRequestElement('action');
		} elseif (isActionSet()) {
			// Get it directly from URL
			return getAction();
		} elseif (($what == 'overview') || (!isWhatSet())) {
			// Default value for admin area
			$ret = 'login';
		}
	} elseif (isActionSet()) {
		// Get it directly from URL
		return getAction();
	}
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): ret=".$ret."<br />");

	// Does the module have a menu?
	if (ifModuleHasMenu($mode)) {
		// Rewriting modules to menu
		$mode = mapModuleToTable($mode);

		// Guest and member menu is 'main' as the default
		if (empty($ret)) $ret = 'main';

		// Load from database
		$result = SQL_QUERY_ESC("SELECT `action` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `what`='%s' LIMIT 1",
			array($mode, $what), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Load action value and pray that this one is the right you want... ;-)
			list($ret) = SQL_FETCHROW($result);
		} // END - if

		// Free memory
		SQL_FREERESULT($result);
	} elseif ((!isExtensionInstalled('sql_patches')) && ($mode != 'admin')) {
		// No sql_patches installed, but maybe we need to register an admin?
		if (isAdminRegistered()) {
			// Redirect to admin area
			redirectToUrl('admin.php');
		} // END - if
	}

	// Return action value
	return $ret;
}

// Get category name back
function getCategory ($cid) {
	// Default is not found
	$ret = getMessage('_CATEGORY_404');

	// Is the category id set?
	if ($cid == 0) {
		// No category
		$ret = getMessage('_CATEGORY_NONE');
	} elseif ($cid > 0) {
		// Lookup the category in database
		$result = SQL_QUERY_ESC("SELECT cat FROM `{?_MYSQL_PREFIX?}_cats` WHERE `id`=%s LIMIT 1",
		array(bigintval($cid)), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Category found... :-)
			list($ret) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Return result
	return $ret;
}

// Get a string of "mail title" and price back
function getPaymentTitlePrice ($pid, $full=false) {
	// Default is not found
	$ret = getMessage('_PAYMENT_404');

	// Load payment data
	$result = SQL_QUERY_ESC("SELECT mail_title, price FROM `{?_MYSQL_PREFIX?}_payments` WHERE `id`=%s LIMIT 1",
	array(bigintval($pid)), __FUNCTION__, __LINE__);
	if (SQL_NUMROWS($result) == 1) {
		// Payment type found... :-)
		if ($full === false) {
			// Return only title
			list($ret) = SQL_FETCHROW($result);
		} else {
			// Return title and price
			list($t, $p) = SQL_FETCHROW($result);
			$ret = $t.' / '.translateComma($p).' {?POINTS?}';
		}
	}

	// Free result
	SQL_FREERESULT($result);

	// Return result
	return $ret;
}

// Get (basicly) the price of given payment id
function getPaymentPoints ($pid, $lookFor = 'price') {
	// Default value...
	$ret = '-1';

	// Search for it in database
	$result = SQL_QUERY_ESC("SELECT %s FROM `{?_MYSQL_PREFIX?}_payments` WHERE `id`=%s LIMIT 1",
	array($lookFor, $pid), __FUNCTION__, __LINE__);

	// Is the entry there?
	if (SQL_NUMROWS($result) == 1) {
		// Payment type found... :-)
		list($ret) = SQL_FETCHROW($result);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return value
	return $ret;
}

// Remove a receiver's ID from $receivers and add a link for him to confirm
function removeReceiver (&$receivers, $key, $userid, $pool_id, $stats_id = '', $bonus = false) {
	// Default is not removed
	$ret = 'failed';

	// Is the userid valid?
	if ($userid > 0) {
		// Remove entry from array
		unset($receivers[$key]);

		// Is there already a line for this user available?
		if ($stats_id > 0) {
			// Only when we got a real stats ID continue searching for the entry
			$type = 'NORMAL'; $rowName = 'stats_id';
			if ($bonus) { $type = 'BONUS'; $rowName = 'bonus_id'; }

			// Try to look the entry up
			$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_user_links` WHERE %s='%s' AND `userid`=%s AND link_type='%s' LIMIT 1",
				array($rowName, $stats_id, bigintval($userid), $type), __FUNCTION__, __LINE__);

			// Was it *not* found?
			if (SQL_NUMROWS($result) == 0) {
				// So we add one!
				SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_user_links` (`%s`, `userid`, `link_type`) VALUES ('%s','%s','%s')",
					array($rowName, $stats_id, bigintval($userid), $type), __FUNCTION__, __LINE__);
				$ret = 'done';
			} else {
				// Already found
				$ret = 'already';
			}

			// Free memory
			SQL_FREERESULT($result);
		}
	}

	// Return status for sending routine
	return $ret;
}

// Calculate sum (default) or count records of given criteria
function countSumTotalData ($search, $tableName, $lookFor = 'id', $whereStatement = 'userid', $countRows = false, $add = '') {
	$ret = 0;
	//* DEBUG: */ outputHtml($search.'/'.$tableName.'/'.$lookFor.'/'.$whereStatement.'/'.$add.'<br />');
	if ((empty($search)) && ($search != '0')) {
		// Count or sum whole table?
		if ($countRows === true) {
			// Count whole table
			$result = SQL_QUERY_ESC("SELECT COUNT(`%s`) FROM `{?_MYSQL_PREFIX?}_%s`".$add,
				array($lookFor, $tableName), __FUNCTION__, __LINE__);
		} else {
			// Sum whole table
			$result = SQL_QUERY_ESC("SELECT SUM(`%s`) FROM `{?_MYSQL_PREFIX?}_%s`".$add,
				array($lookFor, $tableName), __FUNCTION__, __LINE__);
		}
	} elseif (($countRows === true) || ($lookFor == 'userid')) {
		// Count rows
		//* DEBUG: */ outputHtml("COUNT!<br />");
		$result = SQL_QUERY_ESC("SELECT COUNT(`%s`) FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s'".$add,
			array($lookFor, $tableName, $whereStatement, $search), __FUNCTION__, __LINE__);
	} else {
		// Add all rows
		//* DEBUG: */ outputHtml("SUM!<br />");
		$result = SQL_QUERY_ESC("SELECT SUM(`%s`) FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s'".$add,
			array($lookFor, $tableName, $whereStatement, $search), __FUNCTION__, __LINE__);
	}

	// Load row
	list($ret) = SQL_FETCHROW($result);

	// Free result
	SQL_FREERESULT($result);

	// Fix empty values
	if ((empty($ret)) && ($lookFor != 'counter') && ($lookFor != 'id') && ($lookFor != 'userid')) {
		// Float number
		$ret = '0.00000';
	} elseif (''.$ret.'' == '') {
		// Fix empty result
		$ret = 0;
	}

	// Return value
	return $ret;
}
// Getter fro ref level percents
function getReferalLevelPercents ($level) {
	// Default is zero
	$per = 0;

	// Do we have cache?
	if ((isset($GLOBALS['cache_array']['refdepths']['level'])) && (isExtensionActive('cache'))) {
		// First look for level
		$key = array_search($level, $GLOBALS['cache_array']['refdepths']['level']);
		if ($key !== false) {
			// Entry found!
			$per = $GLOBALS['cache_array']['refdepths']['percents'][$key];

			// Count cache hit
			incrementStatsEntry('cache_hits');
		} // END - if
	} elseif (!isExtensionActive('cache')) {
		// Get referal data
		$result_level = SQL_QUERY_ESC("SELECT `percents` FROM `{?_MYSQL_PREFIX?}_refdepths` WHERE `level`='%s' LIMIT 1",
			array(bigintval($level)), __FUNCTION__, __LINE__);

		// Entry found?
		if (SQL_NUMROWS($result_level) == 1) {
			// Get percents
			list($per) = SQL_FETCHROW($result_level);
		} // END - if

		// Free result
		SQL_FREERESULT($result_level);
	}

	// Return percent
	return $per;
}

/**
 *
 * Dynamic referal system, can also send mails!
 *
 * subject     = Subject line, write in lower-case letters and underscore is allowed
 * userid         = Referal ID wich should receive...
 * points      = ... xxx points
 * sendNotify  = shall I send the referal an email or not?
 * rid         = inc/modules/guest/what-confirm.php need this
 * locked      = Shall I pay it to normal (false) or locked (true) points ammount?
 * add_mode    = Add points only to $userid or also refs? (WARNING! Changing 'ref' to 'direct'
 *               for default value will cause no referal will get points ever!!!)
 */
function addPointsThroughReferalSystem ($subject, $userid, $points, $sendNotify = false, $rid = 0, $locked = false, $add_mode = 'ref') {
	//* DEBUG: */ outputHtml("----------------------- <font color=\"#00aa00\">".__FUNCTION__." - ENTRY</font> ------------------------<ul><li>\n");
	// Convert mode to lower-case
	$add_mode = strtolower($add_mode);

	// When $userid = 0 add points to jackpot
	if (($userid == 0) && (isExtensionActive('jackpot'))) {
		// Add points to jackpot
		addPointsToJackpot($points);
		return;
	} // END - if

	// Prepare data for the filter
	$filterData = array(
		'subject'  => $subject,
		'userid'   => $userid,
		'points'   => $points,
		'notify'   => $sendNotify,
		'rid'      => $rid,
		'locked'   => $locked,
		'mode'     => 'add',
		'sub_mode' => $add_mode,
	);

	// Filter it now
	runFilterChain('add_points', $filterData);

	// Count up referal depth
	if (!isset($GLOBALS['ref_level'])) {
		// Initialialize referal system
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): Referal system initialized!<br />");
		$GLOBALS['ref_level'] = 0;
	} else {
		// Increase referal level
		$GLOBALS['ref_level']++;
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): Referal level increased. DEPTH={$GLOBALS['ref_level']}<br />");
	}

	// Default is 'normal' points
	$data = 'points';

	// Which points, locked or normal?
	if ($locked === true) $data = 'locked_points';

	// Check user account
	$result_user = SQL_QUERY_ESC("SELECT `refid`, `email` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s AND `status`='CONFIRMED' LIMIT 1",
		array(bigintval($userid)), __FUNCTION__, __LINE__);

	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},numRows=".SQL_NUMROWS($result_user).",points={$points}<br />");
	if (SQL_NUMROWS($result_user) == 1) {
		// This is the user and his ref
		list($ref, $email) = SQL_FETCHROW($result_user);
		$GLOBALS['cache_array']['add_userid'][$ref] = $userid;

		// Get percents
		$per = getReferalLevelPercents($GLOBALS['ref_level']);
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},points={$points},depth={$GLOBALS['ref_level']},per={$per},mode={$add_mode}<br />");

		// Some percents found?
		if ($per > 0) {
			// Calculate new points
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},points={$points},per={$per},depth={$GLOBALS['ref_level']}<br />");
			$ref_points = $points * $per / 100;

			// Pay refback here if level > 0 and in ref-mode
			if ((isExtensionActive('refback')) && ($GLOBALS['ref_level'] > 0) && ($per < 100) && ($add_mode == "ref") && (isset($GLOBALS['cache_array']['add_userid'][$userid]))) {
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},data={$GLOBALS['cache_array']['add_userid'][$userid]},ref_points={$ref_points},depth={$GLOBALS['ref_level']} - BEFORE!<br />");
				$ref_points = addRefbackPoints($GLOBALS['cache_array']['add_userid'][$userid], $userid, $points, $ref_points);
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},data={$GLOBALS['cache_array']['add_userid'][$userid]},ref_points={$ref_points},depth={$GLOBALS['ref_level']} - AFTER!<br />");
			} // END - if

			// Update points...
			SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_user_points` SET `%s`=`%s`+%s WHERE `userid`=%s AND `ref_depth`='%s' LIMIT 1",
				array($data, $data, $ref_points, bigintval($userid), bigintval($GLOBALS['ref_level'])), __FUNCTION__, __LINE__);
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):data={$data},ref_points={$ref_points},userid={$userid},depth={$GLOBALS['ref_level']},mode={$add_mode} - UPDATE! (".SQL_AFFECTEDROWS().")<br />");

			// No entry updated?
			if (SQL_AFFECTEDROWS() < 1) {
				// First ref in this level! :-)
				SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_user_points` (`userid`,`ref_depth`,`%s`) VALUES (%s,'%s',%s)",
					array($data, bigintval($userid), bigintval($GLOBALS['ref_level']), $ref_points), __FUNCTION__, __LINE__);
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):data={$data},ref_points={$ref_points},userid={$userid},depth={$GLOBALS['ref_level']},mode={$add_mode} - INSERTED! (".SQL_AFFECTEDROWS().")<br />");
			} // END - if

			// Points updated, maybe I shall send him an email?
			if (($sendNotify === true) && ($ref > 0) && ($locked === false)) {
				// Prepare content
				$content = array(
					'percents' => $per,
					'level'    => bigintval($GLOBALS['ref_level']),
					'points'   => $ref_points,
					'refid'    => bigintval($ref)
				);

				// Load email template
				$message = loadEmailTemplate('confirm-referal', $content, bigintval($userid));

				sendEmail($email, THANX_REFERRAL_ONE, $message);
			} elseif (($sendNotify) && ($ref == 0) && ($locked === false) && ($add_mode == 'direct')) {
				// Prepare content
				$content = array(
					'text'   => getMessage('REASON_DIRECT_PAYMENT'),
					'points' => translateComma($ref_points)
				);

				// Load message
				$message = loadEmailTemplate('add-points', $content, $userid);

				// And sent it away
				sendEmail($email, getMessage('SUBJECT_DIRECT_PAYMENT'), $message);
				if (!isGetRequestElementSet('mid')) loadTemplate('admin_settings_saved', false, getMessage('ADMIN_POINTS_ADDED'));
			}

			// Maybe there's another ref?
			if (($ref > 0) && ($points > 0) && ($ref != $userid) && ($add_mode == 'ref')) {
				// Then let's credit him here...
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},ref={$ref},points={$points} - ADVANCE!<br />");
				addPointsThroughReferalSystem(sprintf("%s_ref:%s", $subject, $GLOBALS['ref_level']), $ref, $points, $sendNotify, $ref, $locked);
			} // END - if
		} // END - if
	} // END - if

	// Free result
	SQL_FREERESULT($result_user);
	//* DEBUG: */ outputHtml("</li></ul>----------------------- <font color=\"#aa0000\">".__FUNCTION__." - EXIT</font> ------------------------<br />");
}

// Updates the referal counter
function updateReferalCounter ($userid) {
	// Make it sure referal level zero (member him-/herself) is at least selected
	if (empty($GLOBALS['cache_array']['ref_level'][$userid])) $GLOBALS['cache_array']['ref_level'][$userid] = 1;
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},level={$GLOBALS['cache_array']['ref_level'][$userid]}<br />");

	// Update counter
	SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_refsystem` SET `counter`=`counter`+1 WHERE `userid`=%s AND `level`='%s' LIMIT 1",
		array(bigintval($userid), $GLOBALS['cache_array']['ref_level'][$userid]), __FUNCTION__, __LINE__);

	// When no entry was updated then we have to create it here
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):updated=".SQL_AFFECTEDROWS()."<br />");
	if (SQL_AFFECTEDROWS() < 1) {
		// First count!
		SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_refsystem` (`userid`, `level`, `counter`) VALUES (%s,%s,1)",
			array(bigintval($userid), $GLOBALS['cache_array']['ref_level'][$userid]), __FUNCTION__, __LINE__);
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid}<br />");
	} // END - if

	// Check for his referal
	$result = SQL_QUERY_ESC("SELECT `refid` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
		array(bigintval($userid)), __FUNCTION__, __LINE__);

	// Load refid
	list($ref) = SQL_FETCHROW($result);

	// Free memory
	SQL_FREERESULT($result);
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):userid={$userid},ref={$ref}<br />");

	// When he has a referal...
	if (($ref > 0) && ($ref != $userid)) {
		// Move to next referal level and count his counter one up!
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):ref={$ref} - ADVANCE!<br />");
		$GLOBALS['cache_array']['ref_level'][$userid]++; updateReferalCounter($ref);
	} elseif ((($ref == $userid) || ($ref == 0)) && (getExtensionVersion('cache') >= '0.1.2')) {
		// Remove cache here
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):ref={$ref} - CACHE!<br />");
		rebuildCacheFile('refsystem', 'refsystem');
	}

	// "Walk" back here
	$GLOBALS['cache_array']['ref_level'][$userid]--;

	// Handle refback here if extension is installed
	if (isExtensionActive('refback')) {
		updateRefbackTable($userid);
	} // END - if
}

// Sends out mail to all administrators. This function is no longer obsolete
// because we need it when there is no ext-admins installed
function sendAdminEmails ($subj, $message) {
	// Load all admin email addresses
	$result = SQL_QUERY("SELECT `email` FROM `{?_MYSQL_PREFIX?}_admins` ORDER BY `id` ASC", __FUNCTION__, __LINE__);
	while ($content = SQL_FETCHARRAY($result)) {
		// Send the email out
		sendEmail($content['email'], $subj, $message);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Really simple... ;-)
}

// Get ID number from administrator's login name
function getAdminId ($login) {
	// By default no admin is found
	$ret = '-1';

	// Check cache
	if (isset($GLOBALS['cache_array']['admin']['admin_id'][$login])) {
		// Use it if found to save SQL queries
		$ret = $GLOBALS['cache_array']['admin']['admin_id'][$login];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `login`='%s' LIMIT 1",
			array($login), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			list($ret) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	}
	return $ret;
}

// "Getter" for current admin id
function getCurrentAdminId () {
	// Do we have cache?
	if (!isset($GLOBALS['current_admin_id'])) {
		// Get the admin login from session
		$adminLogin = getSession('admin_login');

		// "Solve" it into an id
		$adminId = getAdminId($adminLogin);

		// Remember in cache securely
		$GLOBALS['current_admin_id'] = bigintval($adminId);
	} // END - if

	// Return it
	return $GLOBALS['current_admin_id'];
}

// Get password hash from administrator's login name
function getAdminHash ($admin) {
	// By default an invalid hash is returned
	$ret = '-1';

	if (isAdminHashSet($admin)) {
		// Check cache
		$ret = $GLOBALS['cache_array']['admin']['password'][$admin];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT password FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array($admin), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Fetch data
			list($ret) = SQL_FETCHROW($result);

			// Set cache
			setAdminHash($admin, $ret);
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	}
	return $ret;
}

// "Getter" for admin login
function getAdminLogin ($adminId) {
	// By default a non-existent login is returned (other functions react on this!)
	$ret = '***';

	if (isset($GLOBALS['cache_array']['admin']['login'][$adminId])) {
		// Get cache
		$ret = $GLOBALS['cache_array']['admin']['login'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = SQL_QUERY_ESC("SELECT login FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
		array(bigintval($adminId)), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) == 1) {
			// Fetch data
			list($ret) = SQL_FETCHROW($result);

			// Set cache
			$GLOBALS['cache_array']['admin']['login'][$adminId] = $ret;
		} // END - if

		// Free memory
		SQL_FREERESULT($result);
	}
	return $ret;
}

// Get email address of admin id
function getAdminEmail ($adminId) {
	// By default an invalid emails is returned
	$ret = '***';

	if (isset($GLOBALS['cache_array']['admin']['email'][$adminId])) {
		// Get cache
		$ret = $GLOBALS['cache_array']['admin']['email'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result_admin_id = SQL_QUERY_ESC("SELECT email FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result_admin_id) == 1) {
			// Get data
			list($ret) = SQL_FETCHROW($result_admin_id);

			// Set cache
			$GLOBALS['cache_array']['admin']['email'][$adminId] = $ret;
		} // END - if

		// Free result
		SQL_FREERESULT($result_admin_id);
	}

	// Return email
	return $ret;
}

// Get default ACL  of admin id
function getAdminDefaultAcl ($adminId) {
	// By default an invalid ACL value is returned
	$ret = '***';

	// Is sql_patches there and was it found in cache?
	if (!isExtensionActive('sql_patches')) {
		// Not found, which is bad, so we need to allow all
		$ret =  'allow';
	} elseif (isset($GLOBALS['cache_array']['admin']['def_acl'][$adminId])) {
		// Use cache
		$ret = $GLOBALS['cache_array']['admin']['def_acl'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result_admin_id = SQL_QUERY_ESC("SELECT `default_acl` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result_admin_id) == 1) {
			// Fetch data
			list($ret) = SQL_FETCHROW($result_admin_id);

			// Set cache
			$GLOBALS['cache_array']['admin']['def_acl'][$adminId] = $ret;
		}

		// Free result
		SQL_FREERESULT($result_admin_id);
	}

	// Return email
	return $ret;
}

// Generates an option list from various parameters
function generateOptionList ($table, $id, $name, $default='', $special='', $where='') {
	$ret = '';
	if ($table == '/ARRAY/') {
		// Selection from array
		if (is_array($id) && is_array($name) && count($id) == count($name)) {
			// Both are arrays
			foreach ($id as $idx => $value) {
				$ret .= '<option value="' . $value . '"';
				if ($default == $value) $ret .= ' selected="selected"';
				$ret .= '>' . $name[$idx] . '</option>';
			} // END - foreach
		} // END - if
	} else {
		// Data from database
		$SPEC = ', '.$id;
		if (!empty($special)) $SPEC = ', ' . $special;
		$ORDER = $name.$SPEC;
		if ($table == 'country') $ORDER = $special;
		$result = SQL_QUERY_ESC("SELECT %s, %s".$SPEC." FROM `{?_MYSQL_PREFIX?}_%s` ".$where." ORDER BY %s",
		array($id, $ORDER, $table, $name), __FUNCTION__, __LINE__);
		if (SQL_NUMROWS($result) > 0) {
			// Found data so add them as OPTION lines: $id is the value and $name is the "name" of the option
			// @TODO Try to rewrite this to $content = SQL_FETCHARRAY()
			while (list($value, $title, $add) = SQL_FETCHROW($result)) {
				if (empty($special)) $add = '';
				$ret .= '<option value="' . $value . '"';
				if ($default == $value) $ret .= ' selected="selected"';
				if (!empty($add)) $add = ' ('.$add.')';
				$ret .= '>' . $title . $add . '</option>';
			} // END - while

			// Free memory
			SQL_FREERESULT($result);
		} else {
			// No data found
			$ret = '<option value="x">{--SELECT_NONE--}</option>';
		}
	}

	// Return - hopefully - the requested data
	return $ret;
}
// Activate exchange
function FILTER_ACTIVATE_EXCHANGE () {
	// Is the extension 'user' there?
	if ((!isExtensionActive('user')) || (getConfig('activate_xchange') == 0)) {
		// Silently abort here
		return false;
	} // END - if

	// Check total amount of users
	$totalUsers = countSumTotalData('CONFIRMED', 'user_data', 'userid', 'status', true, ' AND max_mails > 0');

	if ($totalUsers >= getConfig('activate_xchange')) {
		// Activate System
		setSqlsArray(array(
			"UPDATE `{?_MYSQL_PREFIX?}_mod_reg` SET `locked`='N', `hidden`='N', `mem_only`='Y' WHERE `module`='order' LIMIT 1",
			"UPDATE `{?_MYSQL_PREFIX?}_member_menu` SET `visible`='Y', `locked`='N' WHERE `what`='order' OR `what`='unconfirmed' LIMIT 2",
		));

		// Run SQLs
		runFilterChain('run_sqls');

		// Update configuration
		updateConfiguration('activate_xchange' ,0);

		// Rebuild cache
		rebuildCacheFile('modules', 'modules');
	} // END - if
}

// Deletes a user account with given reason
function deleteUserAccount ($userid, $reason) {
	$points = 0;
	$result = SQL_QUERY_ESC("SELECT (SUM(p.points) - d.used_points) AS points
FROM `{?_MYSQL_PREFIX?}_user_points` AS p
LEFT JOIN `{?_MYSQL_PREFIX?}_user_data` AS d
ON p.userid=d.userid
WHERE p.userid=%s", array(bigintval($userid)), __FUNCTION__, __LINE__);
	if (SQL_NUMROWS($result) == 1) {
		// Save his points to add them to the jackpot
		list($points) = SQL_FETCHROW($result);

		// Delete points entries as well
		SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_points` WHERE `userid`=%s", array(bigintval($userid)), __FUNCTION__, __LINE__);

		// Update mediadata as well
		if (getExtensionVersion('mediadata') >= '0.0.4') {
			// Update database
			updateMediadataEntry(array('total_points'), 'sub', $points);
		} // END - if

		// Now, when we have all his points adds them do the jackpot!
		if (isExtensionActive('jackpot')) addPointsToJackpot($points);
	} // END - if

	// Free the result
	SQL_FREERESULT($result);

	// Delete category selections as well...
	SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_cats` WHERE `userid`=%s",
		array(bigintval($userid)), __FUNCTION__, __LINE__);

	// Remove from rallye if found
	if (isExtensionActive('rallye')) {
		SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_rallye_users` WHERE `userid`=%s",
			array(bigintval($userid)), __FUNCTION__, __LINE__);
	} // END - if

	// Now a mail to the user and that's all...
	$message = loadEmailTemplate('del-user', array('text' => $reason), $userid);
	sendEmail($userid, getMessage('ADMIN_DEL_ACCOUNT'), $message);

	// Ok, delete the account!
	SQL_QUERY_ESC("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1", array(bigintval($userid)), __FUNCTION__, __LINE__);
}

// Generates meta description for given module and 'what' value
function generateMetaDescriptionCode ($mod, $what) {
	// Exclude admin and member's area
	if (($mod != 'admin') && ($mod != 'login')) {
		// Construct dynamic description
		$DESCR = '{?MAIN_TITLE?} '.trim(getConfig('title_middle')) . ' ' . addMenuDescription('guest', 'what-'.$what, true);

		// Output it directly
		outputHtml('<meta name="description" content="' . $DESCR . '" />');
	} // END - if

	// Remove depth
	unset($GLOBALS['ref_level']);
}

// Gets the matching what name from module
function getWhatFromModule ($modCheck) {
	// Is the request element set?
	if (isGetRequestElementSet('what')) {
		// Then return this!
		return getRequestElement('what');
	} // END - if

	// Default is empty
	$what = '';

	//* DEBUG: */ outputHtml(__LINE__.'!'.$modCheck."!<br />");
	switch ($modCheck) {
		case 'admin':
			$what = 'overview';
			break;

		case 'login':
		case 'index':
			$what = getConfig('index_home');
			break;

		default:
			$what = '';
			break;
	} // END - switch

	// Return what value
	return $what;
}

// Subtract points from database and mediadata cache
function subtractPoints ($subject, $userid, $points) {
	// Add points to used points
	SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_user_data` SET `used_points`=`used_points`+%s WHERE `userid`=%s LIMIT 1",
		array($points, bigintval($userid)), __FUNCTION__, __LINE__);

	// Prepare filter data
	$filterData = array(
		'subject' => $subject,
		'userid'  => $userid,
		'points'  => $points,
		'mode'    => 'sub'
	);

	// Insert booking record
	runFilterChain('sub_points', $filterData);
}

// "Getter" for total available receivers
function getTotalReceivers ($mode='normal') {
	// Query database
	$result_all = SQL_QUERY("SELECT
	`userid`
FROM
	`{?_MYSQL_PREFIX?}_user_data`
WHERE
	`status`='CONFIRMED' AND `receive_mails` > 0 ".runFilterChain('exclude_users', $mode),
	__FUNCTION__, __LINE__);

	// Get num rows
	$numRows = SQL_NUMROWS($result_all);

	// Free result
	SQL_FREERESULT($result_all);

	// Return value
	return $numRows;
}

// Returns HTML code with an "<option> list" of all categories
function generateCategoryOptionsList ($mode) {
	// Prepare WHERE statement
	$whereStatement = " WHERE `visible`='Y'";
	if (isAdmin()) $whereStatement = '';

	// Initialize array...
	$CATS = array(
		'id'   => array(),
		'name' => array(),
		'userids' => array()
	);

	// Get categories
	$result = SQL_QUERY("SELECT id, cat FROM `{?_MYSQL_PREFIX?}_cats`".$whereStatement." ORDER BY `sort`", __FUNCTION__, __LINE__);
	if (SQL_NUMROWS($result) > 0) {
		// ... and begin loading stuff
		while ($content = SQL_FETCHARRAY($result)) {
			// Transfer some data
			$CATS['id'][]   = $content['id'];
			$CATS['name'][] = $content['cat'];

			// Check which users are in this category
			$result_userids = SQL_QUERY_ESC("SELECT `userid` FROM `{?_MYSQL_PREFIX?}_user_cats` WHERE `cat_id`=%s",
				array(bigintval($content['id'])), __FUNCTION__, __LINE__);

			// Start adding all
			$userid_cnt = 0;
			// @TODO Rewrite this to $content = SQL_FETCHARRAY()
			while (list($ucat) = SQL_FETCHROW($result_userids)) {
				$result_ver = SQL_QUERY_ESC("SELECT
	`userid`
FROM
	`{?_MYSQL_PREFIX?}_user_data`
WHERE
	`userid`=%s AND `status`='CONFIRMED' AND `receive_mails` > 0".runFilterChain('exclude_users', $mode)."
LIMIT 1",
					array(bigintval($ucat)), __FUNCTION__, __LINE__);

				// Add user count
				$userid_cnt += SQL_NUMROWS($result_ver);

				// Free memory
				SQL_FREERESULT($result_ver);
			} // END - while

			// Free memory
			SQL_FREERESULT($result_userids);

			// Add counter
			$CATS['userids'][] = $userid_cnt;
		} // END - while

		// Free memory
		SQL_FREERESULT($result);

		// Generate options
		$OUT = '';
		foreach ($CATS['id'] as $key => $value) {
			if (strlen($CATS['name'][$key]) > 20) $CATS['name'][$key] = substr($CATS['name'][$key], 0, 17)."...";
			$OUT .= '      <option value="' . $value . '">' . $CATS['name'][$key] . ' (' . $CATS['userids'][$key] . ' {--USER_IN_CAT--})</option>';
		} // END - foreach
	} else {
		// No cateogries are defined yet
		$OUT = '<option class="member_failed">{--MEMBER_NO_CATS--}</option>';
	}

	// Return HTML code
	return $OUT;
}

// Add bonus mail to queue
function addBonusMailToQueue ($subject, $text, $receiverList, $points, $seconds, $url, $cat, $mode='normal', $receiver=0) {
	// Is admin or bonus extension there?
	if (!isAdmin()) {
		// Abort here
		return false;
	} elseif (!isExtensionActive('bonus')) {
		// Abort here
		return false;
	}

	// Calculcate target sent
	$target = countSelection(explode(';', $receiverList));

	// Receiver is zero?
	if ($receiver == 0) {
		// Then auto-fix it
		$receiver = $target;
	} // END - if

	// HTML extension active?
	if (isExtensionActive('html_mail')) {
		// No HTML by default
		$HTML = 'N';

		// HTML mode?
		if ($mode == 'html') $HTML = 'Y';

		// Add HTML mail
		SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_bonus`
(`subject`, `text`, `receivers`, `points`, `time`, `data_type`, `timestamp`, `url`, `cat_id`, `target_send`, `mails_sent`, `html_msg`)
VALUES ('%s','%s','%s','%s','%s','NEW', UNIX_TIMESTAMP(),'%s','%s','%s','%s','%s')",
		array(
			$subject,
			$text,
			$receiverList,
			$points,
			$seconds,
			$url,
			$cat,
			$target,
			bigintval($receiver),
			$HTML
		), __FUNCTION__, __LINE__);
	} else {
		// Add regular mail
		SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_bonus`
(`subject`, `text`, `receivers`, `points`, `time`, `data_type`, `timestamp`, `url`, `cat_id`, `target_send`, `mails_sent`)
VALUES ('%s','%s','%s','%s','%s','NEW', UNIX_TIMESTAMP(),'%s','%s','%s','%s')",
		array(
			$subject,
			$text,
			$receiverList,
			$points,
			$seconds,
			$url,
			$cat,
			$target,
			bigintval($receiver),
		), __FUNCTION__, __LINE__);
	}
}

// Generate a receiver list for given category and maximum receivers
function generateReceiverList ($cat, $receiver, $mode = '') {
	// Init variables
	$CAT_TABS     = "%s";
	$CAT_WHERE    = '';
	$receiverList = '';
	$result       = false;

	// Secure data
	$cat      = bigintval($cat);
	$receiver = bigintval($receiver);

	// Is the receiver zero and mode set?
	if (($receiver == 0) && (!empty($mode))) {
		// Auto-fix receiver maximum
		$receiver = getTotalReceivers($mode);
	} // END - if

	// Category given?
	if ($cat > 0) {
		// Select category
		$CAT_TABS  = "LEFT JOIN `{?_MYSQL_PREFIX?}_user_cats` AS c ON d.userid=c.userid";
		$CAT_WHERE = " AND c.cat_id=%s";
	} // END - if

	// Exclude users in holiday?
	if (getExtensionVersion('holiday') >= '0.1.3') {
		// Add something for the holiday extension
		$CAT_WHERE .= " AND d.`holiday_active`='N'";
	} // END - if

	if ((isExtensionActive('html_mail')) && ($mode == 'html')) {
		// Only include HTML receivers
		$result = SQL_QUERY_ESC("SELECT d.userid FROM `{?_MYSQL_PREFIX?}_user_data` AS d ".$CAT_TABS." WHERE d.`status`='CONFIRMED' AND d.`html`='Y'".$CAT_WHERE." ORDER BY d.%s %s LIMIT %s",
		array($cat, getConfig('order_select'), getConfig('order_mode'), $receiver), __FUNCTION__, __LINE__);
	} else {
		// Include all
		$result = SQL_QUERY_ESC("SELECT d.userid FROM `{?_MYSQL_PREFIX?}_user_data` AS d ".$CAT_TABS." WHERE d.`status`='CONFIRMED'".$CAT_WHERE." ORDER BY d.%s %s LIMIT %s",
		array($cat, getConfig('order_select'), getConfig('order_mode'), $receiver), __FUNCTION__, __LINE__);
	}

	// Entries found?
	if ((SQL_NUMROWS($result) >= $receiver) && ($receiver > 0)) {
		// Load all entries
		while ($content = SQL_FETCHARRAY($result)) {
			// Add receiver when not empty
			if (!empty($content['userid'])) $receiverList .= $content['userid'] . ';';
		} // END - while

		// Free memory
		SQL_FREERESULT($result);

		// Remove trailing semicolon
		$receiverList = substr($receiverList, 0, -1);
	} // END - if

	// Return list
	return $receiverList;
}

// Get timestamp for given stats type and data
function getTimestampFromUserStats ($type, $data, $userid = 0) {
	// Default timestamp is zero
	$stamp = 0;

	// User id set?
	if ((isUserIdSet()) && ($userid == 0)) {
		$userid = getUserId();
	} // END - if

	// Is the extension installed and updated?
	if ((!isExtensionActive('sql_patches')) || (isExtensionOlder('sql_patches', '0.5.6'))) {
		// Return zero here
		return $stamp;
	} // END - if

	// Try to find the entry
	$result = SQL_QUERY_ESC("SELECT
	UNIX_TIMESTAMP(`inserted`) AS stamp
FROM
	`{?_MYSQL_PREFIX?}_user_stats_data`
WHERE
	`userid`=%s AND `stats_type`='%s' AND `stats_data`='%s'
LIMIT 1",
		array(
			bigintval($userid),
			$type,
			$data
		), __FUNCTION__, __LINE__);

	// Is the entry there?
	if (SQL_NUMROWS($result) == 1) {
		// Get this stamp
		list($stamp) = SQL_FETCHROW($result);
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return stamp
	return $stamp;
}

// Inserts user stats
function insertUserStatsRecord ($userid, $type, $data) {
	// Is the extension installed and updated?
	if ((!isExtensionActive('sql_patches')) || (isExtensionOlder('sql_patches', '0.5.6'))) {
		// Return zero here
		return false;
	} // END - if

	// Does it exist?
	if ((!getTimestampFromUserStats($type, $data, $userid)) && (!is_array($data))) {
		// Then insert it!
		SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_user_stats_data` (`userid`,`stats_type`,`stats_data`) VALUES (%s,'%s','%s')",
			array(bigintval($userid), $type, $data), __FUNCTION__, __LINE__);
	} elseif (is_array($data)) {
		// Invalid data!
		logDebugMessage(__FUNCTION__, __LINE__, "userid={$userid},type={$type},data={".gettype($data).": Invalid statistics data type!");
	}
}

// "Getter" for array for user refs and points in given level
function getUserReferalPoints ($userid, $level) {
	//* DEBUG: */ outputHtml("----------------------- <font color=\"#00aa00\">".__FUNCTION__." - ENTRY</font> ------------------------<ul><li>\n");
	// Default is no refs and no nickname
	$add = '';
	$refs = array();

	// Do we have nickname extension installed?
	if (isExtensionActive('nickname')) {
		$add = ', ud.nickname';
	} // END - if

	// Get refs from database
	$result = SQL_QUERY_ESC("SELECT
	ur.id, ur.refid, ud.status, ud.last_online, ud.mails_confirmed, ud.emails_received".$add."
FROM
	`{?_MYSQL_PREFIX?}_user_refs` AS ur
LEFT JOIN
	`{?_MYSQL_PREFIX?}_user_points` AS up
ON
	ur.refid=up.userid AND ur.level=0
LEFT JOIN
	`{?_MYSQL_PREFIX?}_user_data` AS ud
ON
	ur.refid=ud.userid
WHERE
	ur.userid=%s AND ur.level=%s
ORDER BY ur.refid ASC",
		array(
			bigintval($userid),
			bigintval($level)
		), __FUNCTION__, __LINE__);

	// Are there some entries?
	if (SQL_NUMROWS($result) > 0) {
		// Fetch all entries
		while ($row = SQL_FETCHARRAY($result)) {
			// Get total points of this user
			$row['points'] = countSumTotalData($row['refid'], 'user_points', 'points') - countSumTotalData($row['refid'], 'user_data', 'used_points');

			// Get unconfirmed mails
			$row['unconfirmed']  = countSumTotalData($row['refid'], 'user_links', 'id', 'userid', true);

			// Init clickrate with zero
			$row['clickrate'] = 0;

			// Is at least one mail received?
			if ($row['emails_received'] > 0) {
				// Calculate clickrate
				$row['clickrate'] = ($row['mails_confirmed'] / $row['emails_received'] * 100);
			} // END - if

			// Activity is 'active' by default because if autopurge is not installed
			$row['activity'] = getMessage('MEMBER_ACTIVITY_ACTIVE');

			// Is autopurge installed and the user inactive?
			if ((isExtensionActive('autopurge')) && ((time() - getConfig('ap_inactive_since')) >= $row['last_online']))  {
				// Inactive user!
				$row['activity'] = getMessage('MEMBER_ACTIVITY_INACTIVE');
			} // END - if

			// Remove some entries
			unset($row['mails_confirmed']);
			unset($row['emails_received']);
			unset($row['last_online']);

			// Add row
			$refs[$row['id']] = $row;
		} // END - while
	} // END - if

	// Free result
	SQL_FREERESULT($result);

	// Return result
	//* DEBUG: */ outputHtml("</li></ul>----------------------- <font color=\"#aa0000\">".__FUNCTION__." - EXIT</font> ------------------------<br />");
	return $refs;
}

// Recuce the amount of received emails for the receipients for given email
function reduceRecipientReceivedMails ($column, $id, $count) {
	// Search for mail in database
	$result = SQL_QUERY_ESC("SELECT `userid` FROM `{?_MYSQL_PREFIX?}_user_links` WHERE `%s`=%s ORDER BY `userid` ASC LIMIT %s",
		array($column, bigintval($id), $count), __FUNCTION__, __LINE__);

	// Are there entries?
	if (SQL_NUMROWS($result) > 0) {
		// Now load all userids for one big query!
		// @TODO This can be somehow rewritten
		$UIDs = array();
		while (list($userid) = SQL_FETCHROW($result)) {
			$UIDs[$userid] = $userid;
		} // END - while

		// Now update all user accounts
		SQL_QUERY_ESC("UPDATE `{?_MYSQL_PREFIX?}_user_data` SET `emails_received`=`emails_received`-1 WHERE `userid` IN (%s) LIMIT %s",
			array(implode(',', $UIDs), count($UIDs)), __FUNCTION__, __LINE__);
	} // END - if

	// Free result
	SQL_FREERESULT($result);
}

// Creates a new task
function createNewTask ($subject, $notes, $taskType, $userid = 0, $adminId = 0, $strip = true) {
	// Insert the task data into the database
	SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_task_system` (`assigned_admin`, `userid`, `status`, `task_type`, `subject`, `text`, `task_created`) VALUES (%s,%s,'NEW','%s','%s','%s', UNIX_TIMESTAMP())",
		array($adminId, $userid, $taskType, $subject, smartAddSlashes($notes)), __FUNCTION__, __LINE__, true, $strip);
}

// [EOF]
?>
